# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/02_crspd.ipynb.

# %% ../../nbs/01_wrds/02_crspd.ipynb 3
from __future__ import annotations
from typing import List

import pandas as pd
import numpy as np

import pandasmore as pdm
from . import wrds_api

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'LIBRARY', 'TABLE', 'NAMES_TABLE', 'DELIST_TABLE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR',
           'ENTITY_ID_IN_RAW_DSET', 'ENTITY_ID_IN_CLEAN_DSET', 'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET',
           'list_all_vars', 'default_raw_vars', 'parse_varlist', 'get_raw_data', 'process_raw_data', 'delist_adj_ret',
           'features']

# %% ../../nbs/01_wrds/02_crspd.ipynb 4
PROVIDER = 'Wharton Research Data Services (WRDS)'
URL = 'https://wrds-www.wharton.upenn.edu/pages/get-data/center-research-security-prices-crsp/annual-update/stock-security-files/daily-stock-file/'
LIBRARY = 'crsp'
TABLE = 'dsf'
NAMES_TABLE = 'dsenames'
DELIST_TABLE = 'dsedelist'
FREQ = 'D'
MIN_YEAR = 1925
MAX_YEAR = None
ENTITY_ID_IN_RAW_DSET = 'permno'
ENTITY_ID_IN_CLEAN_DSET = 'permno'
TIME_VAR_IN_RAW_DSET = 'date'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/01_wrds/02_crspd.ipynb 5
def list_all_vars() -> pd.DataFrame:
    "Collects names of all available variables from WRDS `{LIBRARY}.{TABLE}` and `{LIBRARY}.{NAMES_TABLE}`"

    try:
        db = wrds_api.Connection()
        msf = db.describe_table(LIBRARY,TABLE).assign(wrds_library=LIBRARY, wrds_table=TABLE)
        mse = db.describe_table(LIBRARY,NAMES_TABLE).assign(wrds_library=LIBRARY, wrds_table=NAMES_TABLE)
        dlst = db.describe_table(LIBRARY,DELIST_TABLE).assign(wrds_library=LIBRARY, wrds_table=DELIST_TABLE)
    finally:
        db.close()

    return pd.concat([msf, mse, dlst])[['name','type','wrds_library','wrds_table']].copy()

# %% ../../nbs/01_wrds/02_crspd.ipynb 9
def default_raw_vars():
    """Defines default variables used in `get_raw_data` if none are specified."""
    
    return ['permno','permco','date',
            'ret', 'retx', 'shrout', 'prc', 
            'shrcd', 'exchcd',
            'cfacpr', 'cfacshr',
            'dlret','dlstcd','dlstdt']  #'siccd','naics','cusip','ncusip'          

# %% ../../nbs/01_wrds/02_crspd.ipynb 11
def parse_varlist(vars: List[str]=None,
                  required_vars = [],
                  ) -> str:
    """Figures out which `vars` come from the `{LIBRARY}.{TABLE}` table and which come from the `{LIBRARY}.{NAMES_TABLE}` table and adds a. and b. prefixes to variable names to feed into an SQL query"""

    # Get all available variables and add suffixes needed for the SQL query
    suffix_mapping = {TABLE: 'a.', NAMES_TABLE: 'b.', DELIST_TABLE: 'c.'}
    all_avail_vars = list_all_vars().drop_duplicates(subset='name',keep='first').copy()
    all_avail_vars['w_prefix'] = all_avail_vars.apply(lambda row: suffix_mapping[row['wrds_table']] + row['name'] , axis=1)

    if vars == '*': return ','.join(list(all_avail_vars['w_prefix']))
    
    # Add required vars to requested vars
    if vars is None: vars = default_raw_vars()
    vars_to_get =  required_vars + [x for x in list(set(vars)) if x not in required_vars]

    # Validate variables to be downloaded (make sure that they are in the target database)
    invalid_vars = [v for v in vars_to_get if v not in list(all_avail_vars.name)]
    if invalid_vars: raise ValueError(f"These vars are not in the database: {invalid_vars}") 

    # Extract information on which variable comes from which wrds table, so we know what prefix to use
    vars_to_get = pd.DataFrame(vars_to_get, columns=['name'])
    get_these = vars_to_get.merge(all_avail_vars, how = 'left', on = 'name')
        
    return ','.join(list(get_these['w_prefix']))

# %% ../../nbs/01_wrds/02_crspd.ipynb 14
def get_raw_data(
        vars: List[str]=None, # If None, downloads `default_raw_vars`; use '*' to get all available variables
        required_vars = ['permno','date'], # Variables that are always downloaded, regardless `vars` argument
        nrows: int=None,  #Number of rows to download. If None, full dataset will be downloaded             
        start_date: str="01/01/1950",  # Start date in MM/DD/YYYY format
        end_date: str=None,            # End date in MM/DD/YYYY format  
        shrcd_exchcd_filters: bool=True, # If true, keep only observations with shrcd in [10,11] and exchcd in [1,2,3]
) -> pd.DataFrame:
    "Downloads `vars` from `start_date` to `end_date` from WRDS {LIBRARY}.{TABLE}, {LIBRARY}.{NAMES_TABLE} and {LIBRARY}.{DELIST_TABLE}." 

    wrds_api.validate_dates([start_date, end_date])
    varlist_string = parse_varlist(vars, required_vars=required_vars)
    sql_string = f"""SELECT {varlist_string}
                        FROM {LIBRARY}.{TABLE} AS a 
                        LEFT JOIN {LIBRARY}.{NAMES_TABLE} AS b
                            ON a.permno=b.permno AND b.namedt<=a.date AND a.date<=b.nameendt 
                        LEFT JOIN {LIBRARY}.{DELIST_TABLE} as c
                            ON a.permno=c.permno AND date_trunc('month', a.date) = date_trunc('month', c.dlstdt)
                            """
    sql_string += "WHERE 1=1 "
    if shrcd_exchcd_filters: sql_string += "AND shrcd IN (10,11) AND exchcd IN (1,2,3) "
    if start_date is not None: sql_string += r"AND date >= %(start_date)s "
    if end_date is not None: sql_string += r"AND date <= %(end_date)s "
    if nrows is not None: sql_string += r" LIMIT %(nrows)s"

    df = wrds_api.download(sql_string,
                             params={'start_date':start_date, 'end_date':end_date, 'nrows':nrows})
    
    return df 

# %% ../../nbs/01_wrds/02_crspd.ipynb 16
def process_raw_data(
        df: pd.DataFrame=None,  # Must contain `permno` and `date` columns         
        clean_kwargs: dict={},  # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
) -> pd.DataFrame:
    """Applies `pandasmore.setup_panel` to `df`"""

    # Change some variables to categorical
    for col in ['permno','permco','shrcd','exchcd']:
        if col in df.columns:
            df[col] = df[col].astype('Int64').astype('category')

    for col in ['naics','cusip','ncusip']:
        if col in df.columns:
            df[col] = df[col].astype('string').astype('category')

    if 'siccd' in df.columns:
        df['siccd'] = df['siccd'].astype('Int64').astype('string').str.zfill(4).astype('category')

    # Set up panel structure
    df = pdm.setup_panel(df, panel_ids=ENTITY_ID_IN_RAW_DSET, time_var=TIME_VAR_IN_RAW_DSET, freq=FREQ, panel_ids_toint=False, **clean_kwargs)
    return df 

# %% ../../nbs/01_wrds/02_crspd.ipynb 19
def delist_adj_ret(
        df: pd.DataFrame, # Requires `ret`,`exchcd`,`dlret`,`dlstcd`, and `dlstdt` variables
        adj_ret_var: str='ret_adj' # Name of the adjusted return variable created by this function
) -> pd.DataFrame:
    """Adjusts for returns for delisting using Shumway and Warther (1999) and Johnson and Zhao (2007)"""

    df['npdelist'] = (df['dlstcd']==500) | df['dlstcd'].between(520,584)
    df['dlret'] = np.where(df['dlret'].isna() & df['npdelist'] & df['exchcd'].isin([1,2]), -0.35, df['dlret'])
    df['dlret'] = np.where(df['dlret'].isna() & df['npdelist'] & df['exchcd'].isin([3]), -0.55, df['dlret'])
    df['dlret'] = np.where(df['dlret'].notna() & df['dlret'] < -1, -1, df['dlret'])
    df['dlret'] = df['dlret'].fillna(0)

    df[adj_ret_var] = (1 + df.ret) * (1 + df['dlret']) - 1
    df[adj_ret_var] = np.where(df[adj_ret_var].isna() & (df['dlret']!=0), df['dlret'], df[adj_ret_var])
    df = df.drop('npdelist', axis=1) 
    return df

# %% ../../nbs/01_wrds/02_crspd.ipynb 21
def features(
        df: pd.DataFrame,
) -> pd.DataFrame:
    
    out = pd.DataFrame(index=df.index)

    out['ret_adj'] = delist_adj_ret(df, adj_ret_var='ret_adj')[['ret_adj']].copy()
    
    # Note that we are not using trading days below, but calendar days
    #the rrolling method below is not feasible since it creates 30 lags all at once so it blows up the dataset
    #out['lbhret12'] = pdm.rrolling(1+df['ret'], window=30, func='prod', skipna=True) - 1
    #out['retvol12'] = pdm.rrolling(df['ret'], window=30, func='std', skipna=True) 

    return out 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/00_fred/00_fred.ipynb.

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'get_series_info', 'default_raw_vars', 'parse_varlist', 'get_raw_data', 'process_raw_data',
           'search']

# %% ../../nbs/00_fred/00_fred.ipynb 4
from typing import List, Dict
import time

import pandas as pd

import pandasmore as pdm
from .fred_api import Fred

# %% ../../nbs/00_fred/00_fred.ipynb 5
PROVIDER = 'Federal Reserve Economic Data (FRED)'
URL = 'https://fred.stlouisfed.org/'

# %% ../../nbs/00_fred/00_fred.ipynb 6
def get_series_info(series: str, # FRED series name
                    api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
                    ) -> pd.Series:
    """Get metadata from FRED for given `series` from FRED"""
    
    return Fred(api_key=api_key).get_series_info(series)

# %% ../../nbs/00_fred/00_fred.ipynb 8
def default_raw_vars() -> pd.DataFrame:
    """List of FRED series that will be used in `get_raw_data()` if none are specified"""

    return ['TB3MS','DTB3','GS10','DGS10','GS1','DGS1','AAA','BAA','DAAA','DBAA','FEDFUNDS','DFF','CPIAUCSL','CPIAUCNS','INDPRO','IPB50001SQ','UNRATE','GDP','GDPC1','GNP','GNPC96','GDPPOT','USREC','RECPROUSM156N','CFNAI','UMCSENT','MICH','USEPUINDXM','USEPUNEWSINDXM','USEPUINDXD','VIXCLS','VXOCLS']


# %% ../../nbs/00_fred/00_fred.ipynb 10
def parse_varlist(vars: List[str]=None, #list of variables requested by user; if None, will use `default_raw_vars()`
                  api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
                  ) -> Dict[str, list]:
    "Splits `vars` by frequency and returns dict with one pd.DataFrame per frequency"
     
    if vars is None: vars = default_raw_vars()

    # Collect frequency data for all series
    freq_store = []
    info_store = []
    for series in vars:
        info = get_series_info(series, api_key)
        info_store.append(info.to_frame().T)
        freq_store.append(pd.DataFrame({'series': [series], 'freq': [info['frequency_short']]}))

    info_store = pd.concat(info_store, ignore_index=True)
    freq_store = pd.concat(freq_store, ignore_index=True)

    # Split the series by frequency
    out = {'info': info_store}
    frequencies = list(freq_store['freq'].value_counts().index)
    for f in frequencies:
        out[f] = list(freq_store['series'].loc[freq_store.freq==f])

    return out

# %% ../../nbs/00_fred/00_fred.ipynb 16
def get_raw_data(vars: str=None, # FRED series name
               api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
               ) -> Dict[str, pd.DataFrame]: 
    """Retrieves `vars` from FRED, splits them by frequency and returns dict with one pd.DataFrame per frequency"""

    api = Fred(api_key=api_key)

    varlist = parse_varlist(vars)

    out = {'info':varlist['info']}
    varlist.pop('info')
    for freq, vars in varlist.items():
        out[freq] = pd.concat([api.get_series(series).to_frame(name=series) for series in vars], axis=1).dropna(how='all')
        
    return out

# %% ../../nbs/00_fred/00_fred.ipynb 21
def process_raw_data(
        data: dict=None,  # keys are frequency str literals; same format as the output of `get_raw_data`
) -> pd.DataFrame: 
    """Processes data from FRED: cleans the date and sets it as index using `pdm.setup_tseries`"""

    out = {'info': data['info']}
    data.pop('info')  
    for freq, df in data.items():  
        df = df.reset_index().rename({'index':'date'},axis=1)
        df = pdm.setup_tseries(df, freq=freq).drop('date', axis=1)
        out[freq] = df 

    return out

# %% ../../nbs/00_fred/00_fred.ipynb 25
def search(search_text: str=None, # What to search for
              order_by: str='popularity', # How to order search results; try `search_rank` if you don't find what you were looking for
              nr_results: int=10, # How many results to output
              api_key: str=None # FRED api key. If None, will use os.getenv("FRED_API_KEY")
              ) -> pd.DataFrame:
    """Search FRED for a given `search_text`, sort by popularity and return only the first `nr_results`"""

    api = Fred(api_key=api_key)  
    return api.search(search_text, order_by=order_by)\
              .pipe(pdm.order_columns, ['title', 'popularity','frequency_short', 'observation_start', 'observation_end'])\
              .iloc[:nr_results].copy()
              



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/02_papers/hassan_etal_2019.ipynb.

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 4
from __future__ import annotations
import pandas as pd

import pandasmore as pdm
from .. import wrds
from ..fetch_tools import get_text_file_from_url

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'HOST_WEBSITE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR', 'ENTITY_ID_IN_RAW_DSET', 'ENTITY_ID_IN_CLEAN_DSET',
           'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET', 'list_all_vars', 'get_raw_data', 'process_raw_data']

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 5
PROVIDER = 'Tarek A. Hassan, Stephan Hollander, Laurence van Lent, Ahmed Tahoun, 2019'
URL = 'https://www.dropbox.com/s/96xo9f1twlu3525/firmquarter_2022q1.csv?raw=1'
HOST_WEBSITE = 'https://www.firmlevelrisk.com/'
FREQ = 'Q'
MIN_YEAR = 2002
MAX_YEAR = 2022
ENTITY_ID_IN_RAW_DSET = 'gvkey'
ENTITY_ID_IN_CLEAN_DSET = 'permno'
TIME_VAR_IN_RAW_DSET = 'date'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 6
def list_all_vars(url: str=URL,
                  delimiter: str='\t'):
    df = get_text_file_from_url(url, nrows=1, delimiter=delimiter)

    return pd.DataFrame(list(df.columns), columns=['name'])

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 9
def get_raw_data(url: str=URL, 
            nrows: int=None, # How many rows to download. If None, all rows are downloaded
            delimiter: str='\t'
            ) -> pd.DataFrame:
    """Download raw data from `url`"""

    return get_text_file_from_url(url, nrows=nrows, delimiter=delimiter)

# %% ../../nbs/02_papers/hassan_etal_2019.ipynb 12
def process_raw_data(
        df: pd.DataFrame=None, # If None, will download using `download_raw`
        gvkey_permno_link: bool|pd.DataFrame=True, # Whether to download permno or not. If DataFrame, must contain `permno`, `gvkey`, and `Qdate`
) -> pd.DataFrame:
    """Converts `gvkey` to string and applies `pandasmore.setup_panel`. Adds `permno` if `gvkey_permno_link` is not False."""

    df = df.copy()

    df['gvkey'] = df['gvkey'].astype('string').str.zfill(6).astype('category') #prepend 0's up to len 6
    df['date'] = df['date'].astype('string')

    # Format date variable so it can be converted into datetime (as the last day of the quarter)
    year = df['date'].str.slice(0, 4).astype('string')
    quarter = df['date'].str.slice(5, 6).astype('int')

    last_month = (quarter * 3).astype('string').str.zfill(2)
    last_day = last_month.map({'03': '31', '06': '30', '09': '30', '12': '31'})

    df['date'] = year + '-' + last_month + '-' + last_day

    df = pdm.setup_panel(df, panel_ids='gvkey', 
                        time_var='date', freq='Q',
                        panel_ids_toint=False,
                        drop_index_duplicates=True, duplicates_which_keep='last')
    
    if not gvkey_permno_link: return df
    else:    
      if gvkey_permno_link is True: gvkey_permno_link = wrds.linking.gvkey_permno_q()
      gvkey_permno_link['gvkey'] = gvkey_permno_link['gvkey'].astype('string').astype('category') 
      df = df.reset_index().merge(gvkey_permno_link, how='inner', on=['gvkey','Qdate'])
      df['permno'] = df['permno'].astype('Int64').astype('category')
      return pdm.setup_panel(df, panel_ids='permno', dates_processed=True, freq='Q',
                                panel_ids_toint=False,
                                drop_index_duplicates=True, duplicates_which_keep='last')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/04_compq_ccm.ipynb.

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 2
from __future__ import annotations
from typing import List

import pandas as pd
import numpy as np

import pandasmore as pdm
from . import wrds_api

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'LIBRARY', 'TABLE', 'LINK_LIBRARY', 'LINK_TABLE', 'COMPANY_TABLE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR',
           'ENTITY_ID_IN_RAW_DSET', 'ENTITY_ID_IN_CLEAN_DSET', 'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET',
           'list_all_vars', 'default_raw_vars', 'parse_varlist', 'get_raw_data', 'process_raw_data', 'ytd_to_quarterly',
           'features']

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 3
PROVIDER = 'Wharton Research Data Services (WRDS)'
URL = 'https://wrds-www.wharton.upenn.edu/pages/get-data/center-research-security-prices-crsp/annual-update/crspcompustat-merged/fundamentals-quarterly/'
LIBRARY = 'comp'
TABLE = 'fundq'
LINK_LIBRARY = 'crsp'
LINK_TABLE = 'ccmxpf_lnkhist'
COMPANY_TABLE = 'company' #contains some header information that is missing from comp.funda (e.g. sic and naics) 
FREQ = 'Q'
MIN_YEAR = 1961
MAX_YEAR = None
ENTITY_ID_IN_RAW_DSET = 'permno'
ENTITY_ID_IN_CLEAN_DSET = 'permno'
TIME_VAR_IN_RAW_DSET = 'datadate'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 4
def list_all_vars() -> pd.DataFrame:
    "Collects names of all available variables from WRDS f`{LIBRARY}.{TABLE}` and `{LIBRARY}.{COMPANY_TABLE}`."

    try:
        db = wrds_api.Connection()
        funda = db.describe_table(LIBRARY,TABLE).assign(wrds_library=LIBRARY, wrds_table=TABLE)
        fundn = db.describe_table(LIBRARY,COMPANY_TABLE).assign(wrds_library=LIBRARY, wrds_table=COMPANY_TABLE)
    finally:
        db.close()

    return pd.concat([funda,fundn])[['name','type','wrds_library','wrds_table']].copy()

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 7
def default_raw_vars():
    """Defines default variables used in `get_raw_data` if none are specified."""

    return ['datadate', 'gvkey', 'cik', 'cusip', 'fyearq', 'fqtr', 'fyr',
            'sic', 'naics', 'exchg', 'rdq', 'fic',
            'atq', 'req', 'xrdq', 'cheq' ,'saleq','revtq', 'dpq', 'ibq', 'cshoq', 'ceqq', 'seqq', 'txdiq', 'ltq', 
            'txditcq', 'pstkq', 'pstkrq', 'lctq', 'actq', 'piq', 'niq', 'cshprq', 'epsfxq', 
            'opepsq', 'epsfiq' ,'epspiq', 'epspxq' ,'dlttq' ,'dlcq' ,'txtq' ,'xintq' ,'ppegtq' ,
            'ppentq' ,'rectq' ,'invtq' ,'cogsq' ,'xsgaq' ,'ajexq' ,'prccq' ,'capxy' ,'oancfy' ,
            'sstky' ,'prstkcy' ,'dltisy' ,'dltry' ,'dvpq' ,'dvy' ,'sppey' ,'aqcy' , 'fopty', 'scstkcy',
            'wcapq' ,'oibdpq' ,'tstkq' ,'apdedateq' ,'fdateq','cdvcy', 'cheq',
            'intanq','gdwlq','mibq', 'oiadpq','ivaoq','npq','rectrq'
            ]             

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 9
def parse_varlist(vars: List[str]=None,
                  required_vars = [],
                  ) -> str:
    """Figures out which `vars` come from the `{LIBRARY}.{TABLE}` table and which come from the `{LIBRARY}.{COMPANY_TABLE}` table and adds a. and b. prefixes to variable names to feed into an SQL query"""

    # Get all available variables and add suffixes needed for the SQL query
    suffix_mapping = {TABLE: 'a.', COMPANY_TABLE: 'b.', }
    all_avail_vars = list_all_vars().drop_duplicates(subset='name',keep='first').copy()
    all_avail_vars['w_prefix'] = all_avail_vars.apply(lambda row: suffix_mapping[row['wrds_table']] + row['name'] , axis=1)

    if vars == '*': return ','.join(list(all_avail_vars['w_prefix']))
    
    # Add required vars to requested vars
    if vars is None: vars = default_raw_vars()
    vars_to_get =  required_vars + [x for x in list(set(vars)) if x not in required_vars]

    # Validate variables to be downloaded (make sure that they are in the target database)
    invalid_vars = [v for v in vars_to_get if v not in list(all_avail_vars.name)]
    if invalid_vars: raise ValueError(f"These vars are not in the database: {invalid_vars}") 

    # Extract information on which variable comes from which wrds table, so we know what prefix to use
    vars_to_get = pd.DataFrame(vars_to_get, columns=['name'])
    get_these = vars_to_get.merge(all_avail_vars, how = 'left', on = 'name')
        
    return ','.join(list(get_these['w_prefix']))

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 11
def get_raw_data(
        vars: List[str]=None, # If None, downloads `default_raw_vars`; use '*' to get all available variables
        required_vars: List[str]=['gvkey','datadate','fyearq','fqtr','rdq'], #list of variables that will get downloaded, even if not in `vars`
        nrows: int=None, #Number of rows to download. If None, full dataset will be downloaded
        start_date: str=None, # Start date in MM/DD/YYYY format
        end_date: str=None #End date in MM/DD/YYYY format
) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS `{LIBRARY}.{TABLE}` and `{LIBRARY}.{COMPANY_TABLE}`. 
        It also adds `sich` and `naicsh` from the annual table (comp.funda)
    """
 
    wrds_api.validate_dates([start_date, end_date])
    vars = parse_varlist(vars, required_vars=required_vars)

    sql_string=f"""SELECT  {vars}, c.sich, c.naicsh,d.lpermno as permno, d.lpermco as permco, d.liid, d.linkprim as linkprim  
                    FROM {LIBRARY}.{TABLE} as a 
                    LEFT JOIN {LIBRARY}.{COMPANY_TABLE} as b ON a.gvkey = b.gvkey
                    LEFT JOIN (
                        SELECT gvkey, fyear, MAX(datadate) as max_date, sich, naicsh
                        FROM comp.funda
                        GROUP BY gvkey, fyear, sich, naicsh
                    ) as c ON a.gvkey = c.gvkey AND a.fyearq = c.fyear
                    INNER JOIN {LINK_LIBRARY}.{LINK_TABLE} AS d ON a.gvkey = d.gvkey 
                    WHERE a.datadate BETWEEN d.linkdt AND COALESCE(d.linkenddt, CURRENT_DATE)
                            AND d.linktype IN ('LU','LC') AND d.linkprim IN ('P','C')
                            AND  a.indfmt='INDL' AND a.datafmt='STD' AND a.popsrc='D' AND a.consol='C'
                """

    if start_date is not None: sql_string += r" AND a.datadate >= %(start_date)s"
    if end_date is not None: sql_string += r" AND a.datadate <= %(end_date)s"
    if nrows is not None: sql_string += r" LIMIT %(nrows)s"
    
    return wrds_api.download(sql_string,
                             params={'start_date':start_date, 'end_date':end_date, 'nrows':nrows})

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 17
def process_raw_data(
        df: pd.DataFrame=None,  # Must contain `permno` and `datadate` columns   
        clean_kwargs: dict={},  # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
) -> pd.DataFrame:
    """Drops duplicates, cleans up dates and applies `pandasmore.setup_panel` to `df`"""

    # Drop permno-datadate duplicates by retaining the latest fyearq-fqtr combination
    df = df.sort_values(['permno','datadate','fyearq','fqtr']).drop_duplicates(subset=['permno','datadate'], keep='last').copy()

    # Clean up some useful dates (convert rdq to datetime, and extract the fiscal year end date)
    df = df.dropna(subset=['fyearq','fqtr']).copy()
    df['rdq'] = pd.to_datetime(df['rdq'])
    df['dtdate_fiscal'] = pd.to_datetime((df['fyearq'].astype(int).astype(str) + '-' 
                                          + (df['fqtr'].astype(int)*3).astype(str) 
                                          + '-1'), format='%Y-%m-%d'
                                          ) + pd.offsets.MonthEnd(1)
    df['Qdate_fiscal'] = df['dtdate_fiscal'].dt.to_period('Q')

    # Change some variables to categorical
    for col in ['permno','permco']:
        if col in df.columns:
            df[col] = df[col].astype('Int64').astype('category')

    for col in ['gvkey','naics','sic','fic','cik','tic','cusip']:
        if col in df.columns:
            df[col] = df[col].astype('string').astype('category')

    if 'sich' in df.columns:
        df['sich'] = df['sich'].astype('Int64').astype('string').str.zfill(4).astype('category')

    if 'naicsh' in df.columns:
        df['naicsh'] = df['naicsh'].astype('Int64').astype('string').astype('category')

    # Set up panel structure     
    df = pdm.setup_panel(df, panel_ids=ENTITY_ID_IN_RAW_DSET, time_var=TIME_VAR_IN_RAW_DSET, freq=FREQ, 
                         panel_ids_toint=False, **clean_kwargs)
    return pdm.order_columns(df,['datadate','dtdate','dtdate_fiscal','Qdate_fiscal','fyearq','fqtr','rdq']) 

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 20
def ytd_to_quarterly(df: pd.DataFrame=None, 
                     vars: List[str]=['capxy','oancfy','sstky' ,'prstkcy','dltisy','dltry','dvy','sppey','aqcy','fopty','scstkcy'],
                     suffix: str='_q' # Suffix to add to the new quarterly variables
) -> pd.DataFrame:
    """Convert YTD variables to quarterly variables by taking the difference between the current and previous quarter."""

    out = df.reset_index().set_index(['permno','Qdate_fiscal']).sort_index().copy()

    new_vars = []
    for v in vars:
        if v in list(out.columns):  
            new_vars.append(v+suffix)
            out[v+suffix] = np.where(out['fqtr']==1, out[v],out[v] - pdm.lag(out[v]))
        else:
            print(f"Variable {v} not found in the dataset")

    return out.reset_index().set_index(['permno','Qdate'])[new_vars].copy()

# %% ../../nbs/01_wrds/04_compq_ccm.ipynb 22
def features(df: pd.DataFrame=None
             ) -> pd.DataFrame:
    """Computes a set of features from `df`"""
    
    # convert ytd variables to quarterly
    out = ytd_to_quarterly(df, suffix='_q')

    # industry 
    out['sic_full'] = df['sich'].astype('object').fillna(df['sic'].astype('object')).astype('category')
    out['naics_full'] = df['naicsh'].astype('object').fillna(df['naics'].astype('object')).astype('category')

    # size
    out['stock_price'] = np.abs(df['prccq'])
    out['mktcap'] = out['stock_price'] * df['cshoq']
    out['lag_atq'] = pdm.lag(df['atq'])

    # book equity vars
    out['pstkq0'] = df['pstkq'].fillna(0)
    out['pref_stock'] = np.where(df['pstkrq'].isnull(), out['pstkq0'], df['pstkrq'])
    out['shreq'] = np.where(df['seqq'].isnull(), df['ceqq'] + out['pstkq0'], df['seqq'])
    out['shreq'] = np.where(out['shreq'].isnull(), df['atq'] - df['ltq'], out['shreq'])
    out['bookeq'] = out['shreq'] + df['txditcq'].fillna(0) - out['pref_stock']

    # issuance vars
    out['equityiss_tot'] = (pdm.rdiff(out['bookeq']) - pdm.rdiff(df['req'])) 
    out['equityiss_cfs'] = (out['sstky_q'].fillna(0) - out['prstkcy_q'].fillna(0))
    out['debtiss_tot'] = (pdm.rdiff(df['atq']) - pdm.rdiff(out['bookeq'])) 
    out['debtiss_cfs'] = (out['dltisy_q'].fillna(0) - out['dltry_q'].fillna(0)) 
    out['debtiss_bs'] = (pdm.rdiff(df['dlttq']) + pdm.rdiff(df['dlcq'].fillna(0))) 
    for v in ['equityiss_tot','equityiss_cfs','debtiss_tot','debtiss_cfs','debtiss_bs']:
        out[f'{v}_2la'] = out[v] / out['lag_atq']

    # investment vars
    out['ppent_pch'] = pdm.rpct_change(df['ppentq'])
    out['capx_2la'] = out['capxy_q'] / out['lag_atq']
    out['tobinq'] = (df['atq'] - out['bookeq'] + out['mktcap']) / df['atq']

    # profitability vars
    out['roa'] = df['ibq'] / df['atq']

    # cash flow vars
    out['cflow_is'] = (df['ibq']+df['dpq']) 
    out['cflow_cfs'] = out['oancfy_q'] 
    out['cflow_full'] = np.where(df.dtdate.dt.year<1987, out['cflow_is'], out['cflow_cfs'])
    for v in ['cflow_is','cflow_cfs','cflow_full']:
        out[f'{v}_2la'] = out[v] / out['lag_atq']

    # liquidity vars
    out['cash_2a'] = df['cheq'] / df['atq']

    # leverage vars
    out['booklev'] = (df['dlttq'] + df['dlcq']) / df['atq']
    out.loc[out.booklev<0, 'booklev'] = 0
    out.loc[out.booklev>1, 'booklev'] = 1

    # payout vars
    out['dividends_2la'] = (out['dvy_q'].fillna(0)+df['dvpq'].fillna(0)) / out['lag_atq']
    out['repurchases_2la'] = (out['prstkcy_q'].fillna(0) - pdm.rdiff(df['pstkrq']).fillna(0)) / out['lag_atq']
    
    out = out.replace([np.inf, -np.inf], np.nan)
    return out 




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/05_ratios.ipynb.

# %% ../../nbs/01_wrds/05_ratios.ipynb 3
from __future__ import annotations
from typing import List

import pandas as pd

import pandasmore as pdm
from . import wrds_api

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'LIBRARY', 'TABLE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR', 'ENTITY_ID_IN_RAW_DSET',
           'ENTITY_ID_IN_CLEAN_DSET', 'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET', 'list_all_vars', 'get_raw_data',
           'process_raw_data', 'keep_only_ratios']

# %% ../../nbs/01_wrds/05_ratios.ipynb 4
PROVIDER = 'Wharton Research Data Services (WRDS)'
URL = 'https://wrds-www.wharton.upenn.edu/pages/get-data/financial-ratios-suite-wrds/financial-ratios-with-ibes-subscription/financial-ratios-firm-level-ibes/'
LIBRARY = 'wrdsapps_finratio_ibes'
TABLE = 'firm_ratio_ibes'
FREQ = 'M'
MIN_YEAR = 1970
MAX_YEAR = None
ENTITY_ID_IN_RAW_DSET = 'permno'
ENTITY_ID_IN_CLEAN_DSET = 'permno'
TIME_VAR_IN_RAW_DSET = 'public_date'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/01_wrds/05_ratios.ipynb 5
def list_all_vars() -> pd.DataFrame:
    "Collects names of all available variables from WRDS f`{LIBRARY}.{TABLE}`"

    try:
        db = wrds_api.Connection()
        funda = db.describe_table(LIBRARY,TABLE).assign(wrds_library=LIBRARY, wrds_table=TABLE)
    finally:
        db.close()

    return funda[['name','type','wrds_library','wrds_table']]

# %% ../../nbs/01_wrds/05_ratios.ipynb 8
def get_raw_data(vars: List[str]=None, # If None or '*', downloads all variables
             nrows: int=None, #Number of rows to download. If None, full dataset will be downloaded
             start_date: str=None, # Start date in MM/DD/YYYY format
             end_date: str=None #End date in MM/DD/YYYY format
             ) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS `{LIBRARY}.{TABLE}` library"""

    wrds_api.validate_dates([start_date, end_date])
    if vars is None or vars=='*': vars = '*'
    else: vars = ','.join(['public_date','permno'] + [f'{x}' for x in vars if x not in ['public_date', 'permno']])

    sql_string=f"""SELECT {vars} FROM {LIBRARY}.{TABLE} WHERE 1 = 1 """
    if start_date is not None: sql_string += r" AND public_date >= %(start_date)s"
    if end_date is not None: sql_string += r" AND public_date <= %(end_date)s"
    if nrows is not None: sql_string += r" LIMIT %(nrows)s"

    return wrds_api.download(sql_string,
                            params={'start_date':start_date, 'end_date':end_date, 'nrows':nrows})

# %% ../../nbs/01_wrds/05_ratios.ipynb 11
def process_raw_data(
        df: pd.DataFrame=None,  # Must contain `permno` and `datadate` columns         
        clean_kwargs: dict={},  # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
) -> pd.DataFrame:
    """Converts some variables to categorical and applies `pandasmore.setup_panel` to `df`"""

    # Convert some columns to categorical
    if 'permno' in df.columns: df['permno'] = df['permno'].astype('Int64').astype('category')
    
    for col in ['gvkey','ticker','cusip','gsector','gicdesc']:
        if col in df.columns: df[col] = df[col].astype('category')

    for col in df.columns:
        if col.startswith('ffi'):
            if col.endswith('desc'): df[col] = df[col].astype('category')
            else: df[col] = df[col].astype('Int64').astype('category')

    # Set panel structure     
    df = pdm.setup_panel(df, panel_ids=ENTITY_ID_IN_RAW_DSET, time_var=TIME_VAR_IN_RAW_DSET, freq=FREQ, panel_ids_toint=False, **clean_kwargs)
    return df 

# %% ../../nbs/01_wrds/05_ratios.ipynb 14
def keep_only_ratios(
        df: pd.DataFrame
) -> pd.DataFrame:
    
    out = pd.DataFrame(index=df.index)

    not_ratios = r"""be, mktcap, price, dtdate, permno, gvkey, ticker, 
                    cusip, public_date, adate, qdate, gsector, gicdesc,
                """.replace("\n", "").replace(' ','').split(',')
    
    for col in list(df.columns):
        if col not in not_ratios and not col.startswith('ffi'):
            out[col] = df[col].copy()

    return out



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/06_ibes_ltg.ipynb.

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 3
from __future__ import annotations
from typing import List

import pandas as pd

import pandasmore as pdm
from . import wrds_api

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'LIBRARY', 'TABLE', 'LINK_LIBRARY', 'LINK_TABLE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR',
           'ENTITY_ID_IN_RAW_DSET', 'ENTITY_ID_IN_CLEAN_DSET', 'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET',
           'list_all_vars', 'default_raw_vars', 'parse_varlist', 'get_raw_data']

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 4
PROVIDER = 'Refinitiv via WRDS'
URL = 'https://wrds-www.wharton.upenn.edu/pages/get-data/ibes-thomson-reuters/ibes-academic/unadjusted-detail/history/'
LIBRARY = 'ibes'
TABLE = 'detu_epsus'
LINK_LIBRARY = 'wrdsapps_link_crsp_ibes'
LINK_TABLE = 'ibcrsphist'
FREQ = 'M'
MIN_YEAR = 1925
MAX_YEAR = None
ENTITY_ID_IN_RAW_DSET = 'permno'
ENTITY_ID_IN_CLEAN_DSET = 'permno'
TIME_VAR_IN_RAW_DSET = 'date'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 5
def list_all_vars() -> pd.DataFrame:
    "Collects names of all available variables from WRDS f`{LIBRARY}.{TABLE}`"

    try:
        db = wrds_api.Connection()
        funda = db.describe_table(LIBRARY,TABLE).assign(wrds_library=LIBRARY, wrds_table=TABLE)
    finally:
        db.close()

    return funda[['name','type','wrds_library','wrds_table']]

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 8
def default_raw_vars():
    return ['ticker', 'value', 'fpi', 'anndats', 'fpedats', 'revdats', 'actdats', 'estimator', 'analys', 'pdf']

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 9
def parse_varlist(vars: List[str]|str=None, #list of variables requested by user
                  required_vars: List[str] = [], #list of variables that will get downloaded, even if not in `vars`
                  prefix: str='a.', #string to add in front of each variable name when we build the SQL string of variable names
                  ) -> str:
    """Adds required variables to requested variables, validates them, and builds the SQL string with their names"""

    if vars=='*': return f'{prefix}*' 

    # Build full list of variables that will be downloaded
    if vars is None: vars = default_raw_vars()
    if required_vars is None: req_vars = []
    vars =  required_vars + [x for x in vars if x not in required_vars] #in case `vars` already contains some of the required variables

    # Validate variables to be downloaded (make sure that they are in the target database)
    valid_vars = list(list_all_vars().name)
    invalid_vars = [v for v in vars if v not in valid_vars]
    if invalid_vars: raise ValueError(f"These vars are not in the database: {invalid_vars}") 

    return ','.join([f'{prefix}{var_name}' for var_name in vars])

# %% ../../nbs/01_wrds/06_ibes_ltg.ipynb 11
def get_raw_data(vars: List[str]=None, # If None, downloads `default_raw_vars`; `permno`, `ticker`, and `anndats` added by default
            required_vars: List[str] = ['ticker', 'anndats'], #list of variables that will get downloaded, even if not in `vars`
             nrows: int=None, #Number of rows to download. If None, full dataset will be downloaded
             start_date: str=None, # Start date in MM/DD/YYYY format
             end_date: str=None, #End date in MM/DD/YYYY format; if None, defaults to current date
             permno_match_score: tuple=(1,), #accuracy of permno-ibes link. 1-6. 1 is best. use >1 with caution.
             ) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS `ibes.detu_epsus` library and adds PERMNO from CRSP"""

    wrds_api.validate_dates([start_date, end_date])
    vars = parse_varlist(vars, prefix='a.', required_vars=required_vars)

    sql_string=f"""SELECT {vars}, b.permno
                        FROM {LIBRARY}.{TABLE} AS a
                        LEFT JOIN {LINK_LIBRARY}.{LINK_TABLE} AS b
                        ON a.ticker = b.ticker
                        WHERE a.anndats BETWEEN b.sdate AND b.edate
                                AND fpi='0'
                """
    if permno_match_score is not None: sql_string += r" AND score IN %(permno_match_score)s"
    if start_date is not None: sql_string += r" AND anndats >= %(start_date)s"
    if end_date is not None: sql_string += r" AND anndats <= %(end_date)s"
    if nrows is not None: sql_string += r" LIMIT %(nrows)s"

    return wrds_api.download(sql_string,
                             params={'permno_match_score': permno_match_score,
                                 'start_date':start_date, 'end_date':end_date, 'nrows':nrows})



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/fetch_tools.ipynb.

# %% ../nbs/fetch_tools.ipynb 3
from __future__ import annotations
import requests
from io import StringIO

import pandas as pd

# %% auto 0
__all__ = ['get_text_file_from_url']

# %% ../nbs/fetch_tools.ipynb 5
def get_text_file_from_url (url, #Data at this url must be readable with pandas.read_csv
             nrows: int=None, #Get only the first `nrows` from the file. If None, gets the entire file
             delimiter: str=',',
             **pd_read_csv_kwargs,
    ) -> pd.DataFrame:
    "Gets the first `nrows` from the file found at `url`. Data at `url` must be separated by `delimiter` and be readable by pandas.read_csv"
    
    if nrows is not None:
        response = requests.get(url, stream=True)
        response.raise_for_status()

        lines = []
        for i, line in enumerate(response.iter_lines(decode_unicode=True)):
            if i >= nrows: break
            lines.append(line)
        partial_csv = '\n'.join(lines)

        return pd.read_csv(StringIO(partial_csv), delimiter=delimiter, **pd_read_csv_kwargs)

    return pd.read_csv(url, delimiter=delimiter,  **pd_read_csv_kwargs)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_wrds/03_compa.ipynb.

# %% ../../nbs/01_wrds/03_compa.ipynb 2
from __future__ import annotations
from typing import List

import pandas as pd
import numpy as np

import pandasmore as pdm
from . import wrds_api

# %% auto 0
__all__ = ['PROVIDER', 'URL', 'LIBRARY', 'TABLE', 'COMPANY_TABLE', 'FREQ', 'MIN_YEAR', 'MAX_YEAR', 'ENTITY_ID_IN_RAW_DSET',
           'ENTITY_ID_IN_CLEAN_DSET', 'TIME_VAR_IN_RAW_DSET', 'TIME_VAR_IN_CLEAN_DSET', 'list_all_vars',
           'default_raw_vars', 'parse_varlist', 'get_raw_data', 'process_raw_data', 'features']

# %% ../../nbs/01_wrds/03_compa.ipynb 3
PROVIDER = 'Wharton Research Data Services (WRDS)'
URL = 'https://wrds-www.wharton.upenn.edu/pages/get-data/compustat-capital-iq-standard-poors/compustat/north-america-daily/fundamentals-annual/'
LIBRARY = 'comp'
TABLE = 'funda'
COMPANY_TABLE = 'company' #contains some header information that is missing from comp.funda (e.g. sic and naics) 
FREQ = 'A'
MIN_YEAR = 1950
MAX_YEAR = None
ENTITY_ID_IN_RAW_DSET = 'gvkey'
ENTITY_ID_IN_CLEAN_DSET = 'gvkey'
TIME_VAR_IN_RAW_DSET = 'datadate'
TIME_VAR_IN_CLEAN_DSET = f'{FREQ}date'

# %% ../../nbs/01_wrds/03_compa.ipynb 4
def list_all_vars() -> pd.DataFrame:
    "Collects names of all available variables from WRDS f`{LIBRARY}.{TABLE}` and `{LIBRARY}.{COMPANY_TABLE}`."

    try:
        db = wrds_api.Connection()
        funda = db.describe_table(LIBRARY,TABLE).assign(wrds_library=LIBRARY, wrds_table=TABLE)
        fundn = db.describe_table(LIBRARY,COMPANY_TABLE).assign(wrds_library=LIBRARY, wrds_table=COMPANY_TABLE)
    finally:
        db.close()

    return pd.concat([funda,fundn])[['name','type','wrds_library','wrds_table']].copy()

# %% ../../nbs/01_wrds/03_compa.ipynb 7
def default_raw_vars():
    """Defines default variables used in `get_raw_data` if none are specified."""

    return ['datadate', 'gvkey', 'cusip' ,'cik' ,'tic' ,'fyear' ,'fyr' , 'fic',
            'naicsh', 'sich' , 'sic', 'naics', 'exchg',  
            'lt' ,'at' ,'txditc' ,'pstkl' ,'pstkrv' ,'pstk' ,'csho' ,'ajex' , 'rdip',
            'act' ,'dvc' ,'xad','seq' ,'che' ,'lct' ,'dlc' ,'ib' ,'dvp' ,'txdi' ,'dp' ,
            'txp' ,'oancf' ,'ivncf' ,'fincf' ,'dltt' ,'mib','ceq' ,'invt' ,'cogs' , 'revt',
            'sale' ,'capx' ,'xrd' ,'txdb' ,'prcc_f' ,'sstk' ,'prstkc' ,'dltis' ,'dltr' ,'emp' ,
            'dd1' ,'ppegt' ,'ppent' ,'xint' ,'txt' ,'sppe' ,'gdwl' ,'xrent' ,'re' ,'dvpsx_f' ,
            'tstk' ,'wcap' ,'rect' ,'xsga' ,'aqc' ,'oibdp' ,'dpact','ni' ,'ivao' ,'ivst' ,
            'dv' , 'intan' ,'pi' ,'txfo' ,'pifo' ,'xpp' ,'drc' ,'drlt' ,'ap' ,'xacc' ,'itcb']             

# %% ../../nbs/01_wrds/03_compa.ipynb 9
def parse_varlist(vars: List[str]=None,
                  required_vars = [],
                  ) -> str:
    """Figures out which `vars` come from the `{LIBRARY}.{TABLE}` table and which come from the `{LIBRARY}.{COMPANY_TABLE}` table and adds a. and b. prefixes to variable names to feed into an SQL query"""

    # Get all available variables and add suffixes needed for the SQL query
    suffix_mapping = {TABLE: 'a.', COMPANY_TABLE: 'b.'}
    all_avail_vars = list_all_vars().drop_duplicates(subset='name',keep='first').copy()
    all_avail_vars['w_prefix'] = all_avail_vars.apply(lambda row: suffix_mapping[row['wrds_table']] + row['name'] , axis=1)

    if vars == '*': return ','.join(list(all_avail_vars['w_prefix']))
    
    # Add required vars to requested vars
    if vars is None: vars = default_raw_vars()
    vars_to_get =  required_vars + [x for x in list(set(vars)) if x not in required_vars]

    # Validate variables to be downloaded (make sure that they are in the target database)
    invalid_vars = [v for v in vars_to_get if v not in list(all_avail_vars.name)]
    if invalid_vars: raise ValueError(f"These vars are not in the database: {invalid_vars}") 

    # Extract information on which variable comes from which wrds table, so we know what prefix to use
    vars_to_get = pd.DataFrame(vars_to_get, columns=['name'])
    get_these = vars_to_get.merge(all_avail_vars, how = 'left', on = 'name')
        
    return ','.join(list(get_these['w_prefix']))

# %% ../../nbs/01_wrds/03_compa.ipynb 11
def get_raw_data(
        vars: List[str]=None, # If None, downloads `default_raw_vars`; use '*' to get all available variables
        required_vars: List[str]=['gvkey','datadate'], #list of variables that will get downloaded, even if not in `vars`
        nrows: int=None, #Number of rows to download. If None, full dataset will be downloaded
        start_date: str=None, # Start date in MM/DD/YYYY format
        end_date: str=None #End date in MM/DD/YYYY format
) -> pd.DataFrame:
    """Downloads `vars` from `start_date` to `end_date` from WRDS `{LIBRARY}.{TABLE}` library """
 
    wrds_api.validate_dates([start_date, end_date])
    vars = parse_varlist(vars, required_vars=required_vars)

    sql_string=f"""SELECT  {vars}  
                    FROM {LIBRARY}.{TABLE} as a 
                    LEFT JOIN {LIBRARY}.{COMPANY_TABLE} as b ON a.gvkey = b.gvkey
                    WHERE  indfmt='INDL' AND datafmt='STD' AND popsrc='D' AND consol='C'
                """
    if start_date is not None: sql_string += r" AND datadate >= %(start_date)s"
    if end_date is not None: sql_string += r" AND datadate <= %(end_date)s"
    if nrows is not None: sql_string += r" LIMIT %(nrows)s"
    
    return wrds_api.download(sql_string,
                             params={'start_date':start_date, 'end_date':end_date, 'nrows':nrows})

# %% ../../nbs/01_wrds/03_compa.ipynb 19
def process_raw_data(
        df: pd.DataFrame=None,  # Must contain `permno` and `datadate` columns   
        clean_kwargs: dict={},  # Params to pass to `pdm.setup_panel` other than `panel_ids`, `time_var`, and `freq`
) -> pd.DataFrame:
    """Applies `pandasmore.setup_panel` to `df`"""

    # Change some variables to categorical
    for col in ['gvkey','naics','sic','fic','cik','tic','cusip']:
        if col in df.columns:
            df[col] = df[col].astype('string').astype('category')

    if 'sich' in df.columns:
        df['sich'] = df['sich'].astype('Int64').astype('string').str.zfill(4).astype('category')

    if 'naicsh' in df.columns:
        df['naicsh'] = df['naicsh'].astype('Int64').astype('string').astype('category')

    # Set up panel structure
    df = pdm.setup_panel(df, panel_ids=ENTITY_ID_IN_RAW_DSET, time_var=TIME_VAR_IN_RAW_DSET, freq=FREQ, 
                         panel_ids_toint=False,
                         **clean_kwargs)
    return df 

# %% ../../nbs/01_wrds/03_compa.ipynb 22
def features(df: pd.DataFrame=None
             ) -> pd.DataFrame:

    out = pd.DataFrame(index=df.index)

    # industry 
    out['sic_full'] = df['sich'].astype('object').fillna(df['sic'].astype('object')).astype('category')
    out['naics_full'] = df['naicsh'].astype('object').fillna(df['naics'].astype('object')).astype('category')

    # size 
    out['stock_price'] = np.abs(df['prcc_f'])
    out['lag_at'] = pdm.lag(df['at'])
    out['mktcap'] = out['stock_price'] * df['csho']

    # book equity vars
    out['pstk0'] = df['pstk'].fillna(0)
    out['pref_stock'] = np.where(df['pstkrv'].isnull(), df['pstkl'], df['pstkrv'])
    out['pref_stock'] = np.where(out['pref_stock'].isnull(),out['pstk0'], out['pref_stock'])
    out['shreq'] = np.where(df['seq'].isnull(), df['ceq'] + out['pstk0'], df['seq'])
    out['shreq'] = np.where(out['shreq'].isnull(), df['at'] - df['lt'], out['shreq'])
    out['bookeq'] = out['shreq'] + df['txditc'].fillna(0) - out['pref_stock']
    out['bookeq_w_itcb'] = out['bookeq'] + df['itcb'].fillna(0)

    out['tobinq'] = (df['at'] - out['bookeq'] + out['stock_price'] * df['csho']) / df['at']

    # issuance vars
    out['equityiss_tot'] = (pdm.rdiff(out['bookeq']) - pdm.rdiff(df['re'])) 
    out['equityiss_cfs'] = (df['sstk'].fillna(0) - df['prstkc'].fillna(0))
    out['debtiss_tot'] = (pdm.rdiff(df['at']) - pdm.rdiff(out['bookeq'])) 
    out['debtiss_cfs'] = (df['dltis'].fillna(0) - df['dltr'].fillna(0)) 
    out['debtiss_bs'] = (pdm.rdiff(df['dltt']) + pdm.rdiff(df['dlc'].fillna(0))) 
    for v in ['equityiss_tot','equityiss_cfs','debtiss_tot','debtiss_cfs','debtiss_bs']:
        out[f'{v}_2la'] = out[v] / out['lag_at']

    # investment vars
    out['ppent_pch'] = pdm.rpct_change(df['ppent'])
    out['capx_2la'] = df['capx'] / out['lag_at']

    # profitability vars
    out['roa'] = df['ib'] / df['at']

    # cash flow vars
    out['cflow_is'] = df['ib']+df['dp']
    out['cflow_cfs'] = df['oancf'] 
    out['cflow_full'] = np.where(df.dtdate.dt.year<1987, out['cflow_is'], out['cflow_cfs'])
    for v in ['cflow_is','cflow_cfs','cflow_full']:
        out[f'{v}_2la'] = out[v] / out['lag_at']

    # liquidity vars
    out['cash_2a'] = df['che'] / df['at']

    # leverage vars
    out['booklev'] = (df['dltt'] + df['dlc']) / df['at']
    out.loc[out.booklev<0, 'booklev'] = 0
    out.loc[out.booklev>1, 'booklev'] = 1

    # payout vars
    out['dividends_2la'] = (df['dvc'].fillna(0) + df['dvp'].fillna(0)) / out['lag_at']
    out['repurchases_2la'] = (df['prstkc'].fillna(0) - pdm.rdiff(df['pstkrv']).fillna(0)) / out['lag_at']

    out = out.replace([np.inf, -np.inf], np.nan)
    return out 




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/finsets',
                'doc_host': 'https://ionmihai.github.io',
                'git_url': 'https://github.com/ionmihai/finsets',
                'lib_path': 'finsets'},
  'syms': { 'finsets.cli': {'finsets.cli.search': ('cli.html#search', 'finsets/cli.py')},
            'finsets.fetch_tools': { 'finsets.fetch_tools.get_text_file_from_url': ( 'fetch_tools.html#get_text_file_from_url',
                                                                                     'finsets/fetch_tools.py')},
            'finsets.fred.fred': { 'finsets.fred.fred.default_raw_vars': ('00_fred/fred.html#default_raw_vars', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.get_raw_data': ('00_fred/fred.html#get_raw_data', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.get_series_info': ('00_fred/fred.html#get_series_info', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.parse_varlist': ('00_fred/fred.html#parse_varlist', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.process_raw_data': ('00_fred/fred.html#process_raw_data', 'finsets/fred/fred.py'),
                                   'finsets.fred.fred.search': ('00_fred/fred.html#search', 'finsets/fred/fred.py')},
            'finsets.fred.fred_api': { 'finsets.fred.fred_api.Fred': ('00_fred/fred_api.html#fred', 'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__do_series_search': ( '00_fred/fred_api.html#fred.__do_series_search',
                                                                                          'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__fetch_data': ( '00_fred/fred_api.html#fred.__fetch_data',
                                                                                    'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__get_search_results': ( '00_fred/fred_api.html#fred.__get_search_results',
                                                                                            'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.__init__': ( '00_fred/fred_api.html#fred.__init__',
                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred._parse': ( '00_fred/fred_api.html#fred._parse',
                                                                              'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series': ( '00_fred/fred_api.html#fred.get_series',
                                                                                  'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_all_releases': ( '00_fred/fred_api.html#fred.get_series_all_releases',
                                                                                               'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_as_of_date': ( '00_fred/fred_api.html#fred.get_series_as_of_date',
                                                                                             'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_first_release': ( '00_fred/fred_api.html#fred.get_series_first_release',
                                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_info': ( '00_fred/fred_api.html#fred.get_series_info',
                                                                                       'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_latest_release': ( '00_fred/fred_api.html#fred.get_series_latest_release',
                                                                                                 'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.get_series_vintage_dates': ( '00_fred/fred_api.html#fred.get_series_vintage_dates',
                                                                                                'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search': ( '00_fred/fred_api.html#fred.search',
                                                                              'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search_by_category': ( '00_fred/fred_api.html#fred.search_by_category',
                                                                                          'finsets/fred/fred_api.py'),
                                       'finsets.fred.fred_api.Fred.search_by_release': ( '00_fred/fred_api.html#fred.search_by_release',
                                                                                         'finsets/fred/fred_api.py')},
            'finsets.papers.hassan_etal_2019': { 'finsets.papers.hassan_etal_2019.get_raw_data': ( '02_papers/hassan_etal_2019.html#get_raw_data',
                                                                                                   'finsets/papers/hassan_etal_2019.py'),
                                                 'finsets.papers.hassan_etal_2019.list_all_vars': ( '02_papers/hassan_etal_2019.html#list_all_vars',
                                                                                                    'finsets/papers/hassan_etal_2019.py'),
                                                 'finsets.papers.hassan_etal_2019.process_raw_data': ( '02_papers/hassan_etal_2019.html#process_raw_data',
                                                                                                       'finsets/papers/hassan_etal_2019.py')},
            'finsets.wrds.compa': { 'finsets.wrds.compa.default_raw_vars': ('01_wrds/compa.html#default_raw_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.features': ('01_wrds/compa.html#features', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.get_raw_data': ('01_wrds/compa.html#get_raw_data', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.list_all_vars': ('01_wrds/compa.html#list_all_vars', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.parse_varlist': ('01_wrds/compa.html#parse_varlist', 'finsets/wrds/compa.py'),
                                    'finsets.wrds.compa.process_raw_data': ( '01_wrds/compa.html#process_raw_data',
                                                                             'finsets/wrds/compa.py')},
            'finsets.wrds.compa_ccm': { 'finsets.wrds.compa_ccm.default_raw_vars': ( '01_wrds/compa_ccm.html#default_raw_vars',
                                                                                     'finsets/wrds/compa_ccm.py'),
                                        'finsets.wrds.compa_ccm.features': ('01_wrds/compa_ccm.html#features', 'finsets/wrds/compa_ccm.py'),
                                        'finsets.wrds.compa_ccm.get_raw_data': ( '01_wrds/compa_ccm.html#get_raw_data',
                                                                                 'finsets/wrds/compa_ccm.py'),
                                        'finsets.wrds.compa_ccm.list_all_vars': ( '01_wrds/compa_ccm.html#list_all_vars',
                                                                                  'finsets/wrds/compa_ccm.py'),
                                        'finsets.wrds.compa_ccm.parse_varlist': ( '01_wrds/compa_ccm.html#parse_varlist',
                                                                                  'finsets/wrds/compa_ccm.py'),
                                        'finsets.wrds.compa_ccm.process_raw_data': ( '01_wrds/compa_ccm.html#process_raw_data',
                                                                                     'finsets/wrds/compa_ccm.py')},
            'finsets.wrds.compq': { 'finsets.wrds.compq.default_raw_vars': ('01_wrds/compq.html#default_raw_vars', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.features': ('01_wrds/compq.html#features', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.get_raw_data': ('01_wrds/compq.html#get_raw_data', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.list_all_vars': ('01_wrds/compq.html#list_all_vars', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.parse_varlist': ('01_wrds/compq.html#parse_varlist', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.process_raw_data': ('01_wrds/compq.html#process_raw_data', 'finsets/wrds/compq.py'),
                                    'finsets.wrds.compq.ytd_to_quarterly': ( '01_wrds/compq.html#ytd_to_quarterly',
                                                                             'finsets/wrds/compq.py')},
            'finsets.wrds.compq_ccm': { 'finsets.wrds.compq_ccm.default_raw_vars': ( '01_wrds/compq_ccm.html#default_raw_vars',
                                                                                     'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.features': ('01_wrds/compq_ccm.html#features', 'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.get_raw_data': ( '01_wrds/compq_ccm.html#get_raw_data',
                                                                                 'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.list_all_vars': ( '01_wrds/compq_ccm.html#list_all_vars',
                                                                                  'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.parse_varlist': ( '01_wrds/compq_ccm.html#parse_varlist',
                                                                                  'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.process_raw_data': ( '01_wrds/compq_ccm.html#process_raw_data',
                                                                                     'finsets/wrds/compq_ccm.py'),
                                        'finsets.wrds.compq_ccm.ytd_to_quarterly': ( '01_wrds/compq_ccm.html#ytd_to_quarterly',
                                                                                     'finsets/wrds/compq_ccm.py')},
            'finsets.wrds.crspd': { 'finsets.wrds.crspd.default_raw_vars': ('01_wrds/crspd.html#default_raw_vars', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.delist_adj_ret': ('01_wrds/crspd.html#delist_adj_ret', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.features': ('01_wrds/crspd.html#features', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.get_raw_data': ('01_wrds/crspd.html#get_raw_data', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.list_all_vars': ('01_wrds/crspd.html#list_all_vars', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.parse_varlist': ('01_wrds/crspd.html#parse_varlist', 'finsets/wrds/crspd.py'),
                                    'finsets.wrds.crspd.process_raw_data': ( '01_wrds/crspd.html#process_raw_data',
                                                                             'finsets/wrds/crspd.py')},
            'finsets.wrds.crspm': { 'finsets.wrds.crspm.default_raw_vars': ('01_wrds/crspm.html#default_raw_vars', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.delist_adj_ret': ('01_wrds/crspm.html#delist_adj_ret', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.features': ('01_wrds/crspm.html#features', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.get_raw_data': ('01_wrds/crspm.html#get_raw_data', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.list_all_vars': ('01_wrds/crspm.html#list_all_vars', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.parse_varlist': ('01_wrds/crspm.html#parse_varlist', 'finsets/wrds/crspm.py'),
                                    'finsets.wrds.crspm.process_raw_data': ( '01_wrds/crspm.html#process_raw_data',
                                                                             'finsets/wrds/crspm.py')},
            'finsets.wrds.ibes_ltg': { 'finsets.wrds.ibes_ltg.default_raw_vars': ( '01_wrds/ibes_ltg.html#default_raw_vars',
                                                                                   'finsets/wrds/ibes_ltg.py'),
                                       'finsets.wrds.ibes_ltg.get_raw_data': ( '01_wrds/ibes_ltg.html#get_raw_data',
                                                                               'finsets/wrds/ibes_ltg.py'),
                                       'finsets.wrds.ibes_ltg.list_all_vars': ( '01_wrds/ibes_ltg.html#list_all_vars',
                                                                                'finsets/wrds/ibes_ltg.py'),
                                       'finsets.wrds.ibes_ltg.parse_varlist': ( '01_wrds/ibes_ltg.html#parse_varlist',
                                                                                'finsets/wrds/ibes_ltg.py')},
            'finsets.wrds.linking': { 'finsets.wrds.linking.gvkey_permno_a': ( '01_wrds/linking.html#gvkey_permno_a',
                                                                               'finsets/wrds/linking.py'),
                                      'finsets.wrds.linking.gvkey_permno_m': ( '01_wrds/linking.html#gvkey_permno_m',
                                                                               'finsets/wrds/linking.py'),
                                      'finsets.wrds.linking.gvkey_permno_q': ( '01_wrds/linking.html#gvkey_permno_q',
                                                                               'finsets/wrds/linking.py'),
                                      'finsets.wrds.linking.ibes_ticker_permno': ( '01_wrds/linking.html#ibes_ticker_permno',
                                                                                   'finsets/wrds/linking.py')},
            'finsets.wrds.ratios': { 'finsets.wrds.ratios.get_raw_data': ('01_wrds/ratios.html#get_raw_data', 'finsets/wrds/ratios.py'),
                                     'finsets.wrds.ratios.keep_only_ratios': ( '01_wrds/ratios.html#keep_only_ratios',
                                                                               'finsets/wrds/ratios.py'),
                                     'finsets.wrds.ratios.list_all_vars': ('01_wrds/ratios.html#list_all_vars', 'finsets/wrds/ratios.py'),
                                     'finsets.wrds.ratios.process_raw_data': ( '01_wrds/ratios.html#process_raw_data',
                                                                               'finsets/wrds/ratios.py')},
            'finsets.wrds.wrds_api': { 'finsets.wrds.wrds_api.Connection': ('01_wrds/wrds_api.html#connection', 'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__check_schema_perms': ( '01_wrds/wrds_api.html#connection.__check_schema_perms',
                                                                                                  'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__create_pgpass_file_unix': ( '01_wrds/wrds_api.html#connection.__create_pgpass_file_unix',
                                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__create_pgpass_file_win32': ( '01_wrds/wrds_api.html#connection.__create_pgpass_file_win32',
                                                                                                        'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__enter__': ( '01_wrds/wrds_api.html#connection.__enter__',
                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__exit__': ( '01_wrds/wrds_api.html#connection.__exit__',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__get_schema_for_view': ( '01_wrds/wrds_api.html#connection.__get_schema_for_view',
                                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__get_user_credentials': ( '01_wrds/wrds_api.html#connection.__get_user_credentials',
                                                                                                    'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__init__': ( '01_wrds/wrds_api.html#connection.__init__',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__make_sa_engine_conn': ( '01_wrds/wrds_api.html#connection.__make_sa_engine_conn',
                                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.__write_pgpass_file': ( '01_wrds/wrds_api.html#connection.__write_pgpass_file',
                                                                                                 'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.close': ( '01_wrds/wrds_api.html#connection.close',
                                                                                   'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.connect': ( '01_wrds/wrds_api.html#connection.connect',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.create_pgpass_file': ( '01_wrds/wrds_api.html#connection.create_pgpass_file',
                                                                                                'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.describe_table': ( '01_wrds/wrds_api.html#connection.describe_table',
                                                                                            'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.get_row_count': ( '01_wrds/wrds_api.html#connection.get_row_count',
                                                                                           'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.get_table': ( '01_wrds/wrds_api.html#connection.get_table',
                                                                                       'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.list_libraries': ( '01_wrds/wrds_api.html#connection.list_libraries',
                                                                                            'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.list_tables': ( '01_wrds/wrds_api.html#connection.list_tables',
                                                                                         'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.load_library_list': ( '01_wrds/wrds_api.html#connection.load_library_list',
                                                                                               'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.Connection.raw_sql': ( '01_wrds/wrds_api.html#connection.raw_sql',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.NotSubscribedError': ( '01_wrds/wrds_api.html#notsubscribederror',
                                                                                     'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.SchemaNotFoundError': ( '01_wrds/wrds_api.html#schemanotfounderror',
                                                                                      'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.download': ('01_wrds/wrds_api.html#download', 'finsets/wrds/wrds_api.py'),
                                       'finsets.wrds.wrds_api.validate_dates': ( '01_wrds/wrds_api.html#validate_dates',
                                                                                 'finsets/wrds/wrds_api.py')}}}
